/*
+-----------------------------------------------------------------------+
|                                                                       |
|          Program Name: ParSer                                         |
|                Author: Michael Krmer                                 |
|                        dk8pp@ish.de                 (private)         |
|                        Michael.Kraemer@eu.necel.com (business)        |
|                  Date: 27-Jan-2005                                    |
|              Language: Open Watcom C 1.5                              |
|               Version: 0.93                                           |
|        Program status: Initial release, unoptimized brute force search|
|                                                                       |
+-----------------------------------------------------------------------+
|                                                                       |
|  Purpose: Find optimum set of resistors or capacitors for a given     |
|           total resistance or capacitance.                            |
|                                                                       |
|  To do for next version:                                              |
|  - Terminate individual search loops, if values exceed the maximum    |
|    deviation                                                          |
|  - Terminate loop if target cannot be achieved with current value     |
|                                                                       |
+-----------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------+
|       Includes                                                        |
+-----------------------------------------------------------------------+
*/
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <math.h>
#include <time.h>
#include "WinDef.h"
#include "mytype.h"
#include "ParSer.h"

/*
+-----------------------------------------------------------------------+
|       Defines                                                         |
+-----------------------------------------------------------------------+
*/
#define MAX_LOADSTRING 100      /* maximum length of strings */
#define MAX_TARGET 100E6        /* maximum target value */
#define RDECADES 8              /* number of decades for resistors */
#define CDECADES 10             /* number of decades for capacitors */
#define MAXDECADES 10           /* max(RDECADES,CDECADES) */
#define NOS 500                 /* maximum number of solutions */

#define RESIZE_X 0x1            /* resize control in x direction */
#define RESIZE_Y 0x2            /* resize control in y direction */
#define MOVE_X   0x4            /* move control in x direction */
#define MOVE_Y   0x8            /* move control in y direction */

/* private messages */
#define USER_PRINTRESULTS 100   /* output results to the screen */
#define USER_PRINTCOUNT   101   /* print number of found solutions */

#define MIN_SIZE_X 380          /* Minimum horizontal dialog size */
#define MIN_SIZE_Y 420          /* Minimum vertical dialog size */

/*
+-----------------------------------------------------------------------+
       structs and enums                                               
+-----------------------------------------------------------------------+
*/

typedef enum {S, SS, PP, PPP, SPP, SSS, SSP, SSS1, SSS2, PPS} CONF ;
/* CONF is the configuration of the network.
   S:    1 single R/C
   SS:   2 serial R/C
   PP:   2 parallel R/C
   PPP:  3 parallel R/C
   SPP:  1 serial R/C, 2 parallel R/C
   SSS:  3 serial R/C
   SSP:  2 serial R/C, 1 parallel R/C

the following definitions are for resistor networks only
   SSS1: 3 serial R, tap between first two resistors
   SSS2: 3 serial R, tap between last two resistors
   PPS:  2 parallel R, 1 serial R               */

typedef enum {R, C, RN} COMP ;

typedef struct {_DOUBLE target ;
                _DOUBLE res ;
                _DOUBLE val1 ;
                _DOUBLE val2 ;
                _DOUBLE val3 ;
                _DOUBLE err ;
                CONF conf ;
                BOOL valid ;} SOLUTION ;

typedef struct {
        _WORD control ;
        _WORD left ;
        _WORD top ;
        _WORD right ;
        _WORD bottom ;
        _WORD flags ;
                } COORDS ;

/*
+-----------------------------------------------------------------------+
|       global variables                                                |
+-----------------------------------------------------------------------+
*/
HINSTANCE       MyInstance ;
HWND            MainWindow ;
unsigned        ScreenWidthInMM;
unsigned        ScreenHeightInMM;
_BYTE buffer[256] ; // general purpose buffer
_BYTE static strng[MAX_LOADSTRING] ;
_BYTE static strng1[MAX_LOADSTRING] ;
BOOL TARGET_error = FALSE ;
BOOL MAXERR_error = FALSE ;
BOOL MAXDEVIATION_error = FALSE ;
BOOL valid_imin, valid_imax ;
_DOUBLE maxerror = 0.1 ;
_DOUBLE maxdeviation = 10 ;
_DOUBLE res, imin, imax, iload ;
COMP component ;

_DOUBLE e6[] =   {100, 150, 220, 330, 470, 680} ;

/* e12 is a superset of e6 */
_DOUBLE e12[] =  {100, 120, 150, 180, 220, 270, 330, 390, 470, 560, 680, 820} ;

/* e24 is a superset of e12 */
_DOUBLE e24[] =  {100, 110, 120, 130, 150, 160, 180, 200, 220, 240, 270, 300,
                  330, 360, 390, 430, 470, 510, 560, 620, 680, 750, 820, 910} ;

/* Note that e48 is NOT a superset of e24 */
_DOUBLE e48[] =  {100, 105, 110, 115, 121, 127, 133, 140, 147, 154, 162, 169,
                  178, 187, 196, 205, 215, 226, 237, 249, 261, 274, 287, 301,
                  316, 332, 348, 365, 383, 402, 422, 442, 464, 487, 511, 536,
                  562, 590, 619, 649, 681, 715, 750, 787, 825, 866, 909, 953} ;

/* e96 is a superset of e48 */
_DOUBLE e96[] =  {100, 102, 105, 107, 110, 113, 115, 118, 121, 124, 127, 130,
                  133, 137, 140, 143, 147, 150, 154, 158, 162, 165, 169, 174,
                  178, 182, 187, 191, 196, 200, 205, 210, 215, 221, 226, 232,
                  237, 243, 249, 255, 261, 267, 274, 280, 287, 294, 301, 309,
                  316, 324, 332, 340, 348, 357, 365, 374, 383, 392, 402, 412,
                  422, 432, 442, 453, 464, 475, 487, 499, 511, 523, 536, 549,
                  562, 576, 590, 604, 619, 634, 649, 665, 681, 698, 715, 732,
                  750, 768, 787, 806, 825, 845, 866, 887, 909, 931, 953, 976 } ;

/* e192 is a superset of e96 */
_DOUBLE e192[] = {100, 101, 102, 104, 105, 106, 107, 109, 110, 111, 113, 114,
                  115, 117, 118, 120, 121, 123, 124, 126, 127, 129, 130, 132,
                  133, 135, 137, 138, 140, 142, 143, 145, 147, 149, 150, 152,
                  154, 156, 158, 160, 162, 164, 165, 167, 169, 172, 174, 176,
                  178, 180, 182, 184, 187, 189, 191, 193, 196, 198, 200, 203,
                  205, 208, 210, 213, 215, 218, 221, 223, 226, 229, 232, 234,
                  237, 240, 243, 246, 249, 252, 255, 258, 261, 264, 267, 271,
                  274, 277, 280, 284, 287, 291, 294, 298, 301, 305, 309, 312,
                  316, 320, 324, 328, 332, 336, 340, 344, 348, 352, 357, 361,
                  365, 370, 374, 379, 383, 388, 392, 397, 402, 407, 412, 417,
                  422, 427, 432, 437, 442, 448, 453, 459, 464, 470, 475, 481,
                  487, 493, 499, 505, 511, 517, 523, 530, 536, 542, 549, 556,
                  562, 569, 576, 583, 590, 597, 604, 612, 619, 626, 634, 642,
                  649, 657, 665, 673, 681, 690, 698, 706, 715, 723, 732, 741,
                  750, 759, 768, 777, 787, 796, 806, 816, 825, 835, 845, 856,
                  866, 876, 887, 898, 909, 920, 931, 942, 953, 965, 976, 988 } ;

// reserve space for decades
_DOUBLE table[MAXDECADES * (sizeof(e192)+sizeof(e24))/sizeof(e192[0])] ; /* e24 contains some unique values */
_WORD tablesize ;
SOLUTION sol[NOS] ;
_DOUBLE target, target2 ;
_BOOL topos, topopp, toposs, topoppp, toposss, topospp, topossp ;
_BOOL volatile abort_thread ;
_WORD volatile counter = 0 ;
_BOOL status[40] ; /* stores enabled/disabled status of controls (make sufficiently large...) */

COORDS coordinates[] = {{0, 0, 0, 0, 0, 0},  /* first entry is reserved for main dialog */
                        {DLG_RESULT, 0, 0, 0, 0, RESIZE_X|RESIZE_Y},
                        {DLG_STATUS, 0, 0, 0, 0, RESIZE_X|RESIZE_Y}
                       } ;

/*
+-----------------------------------------------------------------------+
|       Function prototypes                                             |
+-----------------------------------------------------------------------+
*/
BOOL CALLBACK ParSerProc(HWND window_handle, unsigned msg,
                                WPARAM wparam, LPARAM lparam ) ;
BOOL get_double_value(_BYTE *buffer, _DOUBLE *value) ;
void fill_table(void) ;
_WORD mix_table(_DOUBLE *target, _DOUBLE *source1, _WORD size1,
                                 _DOUBLE *source2, _WORD size2) ;
void enqueue(_DOUBLE target, _DOUBLE res, CONF conf,
             _DOUBLE val1, _DOUBLE val2, _DOUBLE val3) ;
void print_results(void) ;
_BYTE *pval(_BYTE *strng, _DOUBLE val) ;
DWORD WINAPI FindResults(LPVOID lpParam) ;
void EnableControls(_BOOL enable) ;
void SaveControls(_BOOL *status) ;
void RestoreControls(_BOOL *status) ;
void get_coords(COORDS *coordinates, _WORD size) ;
void resize_controls(COORDS *coordinates, _WORD size) ;

/*
+-----------------------------------------------------------------------+
|       WinMain                                                         |
+-----------------------------------------------------------------------+
|                                                                       |
| MAIN program.                                                         |
|                                                                       |
+-----------------------------------------------------------------------+
*/
int WINAPI WinMain(HINSTANCE this_inst, HINSTANCE prev_inst,
                   LPSTR cmdline, int cmdshow )
{
//WNDCLASSEX WndClsX ;
//_WORD static volatile retcd ;
_LONG static volatile rv ;
   SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS) ; /* reduce priority */
   MyInstance = this_inst ;
   InitCommonControls() ;
   rv = DialogBox(this_inst, "DLG_PARSER", 0, (DLGPROC)ParSerProc );
   return( 0 );
}

/*
+-----------------------------------------------------------------------+
|       ParSerProc                                                      |
+-----------------------------------------------------------------------+
|                                                                       |
| Processes messages for the DLG_PARSER dialog                          |
|                                                                       |
+-----------------------------------------------------------------------+
*/
BOOL CALLBACK ParSerProc(HWND window_handle, unsigned msg,
                         WPARAM wparam, LPARAM lparam)
{
_WORD i ;
_DOUBLE input_value ;
HANDLE icons, iconss, iconsss, iconpp, iconppp, iconspp, iconssp ;
//_WORD static volatile retcd ;
DWORD static dwThreadId, dwThrdParam = 1;
HANDLE static hThread;
HWND msg_hwnd ;
_BOOL static suspended ;
BOOL retcd ;
RECT static rect ;
clock_t static start_time, elapsed_time ;
HFONT hFont ;
WPARAM fwSide ;
LPRECT lprc ;

   retcd = FALSE ; /* return FALSE if we don't process the message */

   switch(msg)
   {
   case WM_INITDIALOG:
      retcd = TRUE ;
   /* get the controls in the dialog initialized with the correct
      settings based on the control data */
      MainWindow = window_handle ;

// save original coordinates of dialog
      GetWindowRect(MainWindow, &rect) ;
      coordinates[0].top = rect.top ;
      coordinates[0].right = rect.right ;
      coordinates[0].bottom = rect.bottom ;
      coordinates[0].left = rect.left ;

// save original coordinates of controls
      get_coords(&coordinates[0], sizeof(coordinates)/sizeof(coordinates[0])) ;

      CheckDlgButton(window_handle, DLG_TOPOS, BST_CHECKED) ;
      CheckDlgButton(window_handle, DLG_TOPOSS, BST_CHECKED) ;
      CheckDlgButton(window_handle, DLG_TOPOPP, BST_CHECKED) ;
      SetDlgItemText(MainWindow, DLG_MAXERROR, "1 %") ;
      SetDlgItemText(MainWindow, DLG_MAXDEVIATION, "10") ;
      CheckRadioButton(window_handle,RAD_COMPR,RAD_COMPRN,RAD_COMPR) ;
      SendMessage(window_handle, WM_COMMAND, RAD_COMPR, NULL);
      SendMessage(window_handle, WM_SETICON, ICON_BIG, (LPARAM) LoadIcon(MyInstance, "IC_A_DEFAULT1"));

// Select font for controls
      hFont = CreateFont(
          15,                  // logical height of font
          0,                   // logical average character width
          0,                   // angle of escapement
          0,                   // base-line orientation angle
          FW_NORMAL,           // font weight
          FALSE,               // italic attribute flag
          FALSE,               // underline attribute flag
          FALSE,               // strikeout attribute flag
          ANSI_CHARSET,        // character set identifier
          OUT_DEFAULT_PRECIS,  // output precision
          CLIP_DEFAULT_PRECIS, // clipping precision
          DEFAULT_QUALITY,     // output quality
          DEFAULT_PITCH,       // pitch and family
          "Courier New"        // pointer to typeface name string
      ) ;

      SendMessage(GetDlgItem(MainWindow, DLG_RESULT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_IMINTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_IMAXTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_ILOADTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_TARGETTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_TARGETTEXT2), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_MAXDEVIATIONTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
      SendMessage(GetDlgItem(MainWindow, DLG_MAXERRORTEXT), WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));

   break ;

#if CLOCKS_PER_SEC != 1000
#error CLOCKS_PER_SEC should be 1000. Please adapt the source code accordingly.
#endif
   case WM_USER:
      retcd = TRUE ;
      switch(LOWORD(wparam))
      {
      case USER_PRINTRESULTS:
      // output the results
         elapsed_time = clock() - start_time ;
         print_results() ;
         if (abort_thread)
            LoadString(MyInstance, STR_ABORTED, strng, MAX_LOADSTRING);
         else
         {
            LoadString(MyInstance, STR_SOLUTIONS, strng1, MAX_LOADSTRING);
            sprintf(strng, strng1, counter, (elapsed_time/CLOCKS_PER_SEC)/60,
                   (elapsed_time/CLOCKS_PER_SEC)%60, elapsed_time%CLOCKS_PER_SEC) ;
         }
         SetDlgItemText(MainWindow, DLG_STATUS, strng) ;
         CloseHandle(hThread);
         hThread = NULL ;
         SetDlgItemText(MainWindow, DLG_START, "Start") ;
         RestoreControls(status) ; /* restore controls */
         EnableWindow(GetDlgItem(MainWindow, DLG_SUSPEND), FALSE) ;
      break ;

      case USER_PRINTCOUNT:
      // display number of solutions
         LoadString(MyInstance, STR_SOLUTIONSSOFAR, strng, MAX_LOADSTRING);
         sprintf(strng1, "%u %s", counter, strng) ;
         SetDlgItemText(MainWindow, DLG_STATUS, strng1) ;
      break ;
      }
   break ;

   case WM_SIZE:
      retcd = 0 ;
      resize_controls(&coordinates[0], sizeof(coordinates)/sizeof(coordinates[0])) ;
   break ;

   case WM_SIZING:
      retcd = TRUE ;
      fwSide = wparam;         // edge of window being sized
      lprc = (LPRECT) lparam;  // screen coordinates of drag rectangle
      if ((fwSide == WMSZ_RIGHT) || (fwSide == WMSZ_BOTTOMRIGHT) || (fwSide == WMSZ_TOPRIGHT))
         if ((lprc->right-lprc->left) < MIN_SIZE_X) lprc->right = lprc->left + MIN_SIZE_X ;
      if ((fwSide == WMSZ_LEFT) || (fwSide == WMSZ_BOTTOMLEFT) || (fwSide == WMSZ_TOPLEFT))
         if ((lprc->right-lprc->left) < MIN_SIZE_X) lprc->left = lprc->right - MIN_SIZE_X ;
      if ((fwSide == WMSZ_BOTTOM) || (fwSide == WMSZ_BOTTOMRIGHT) || (fwSide == WMSZ_BOTTOMLEFT))
         if ((lprc->bottom-lprc->top) < MIN_SIZE_Y) lprc->bottom = lprc->top + MIN_SIZE_Y ;
      if ((fwSide == WMSZ_TOP) || (fwSide == WMSZ_TOPRIGHT) || (fwSide == WMSZ_TOPLEFT))
         if ((lprc->bottom-lprc->top) < MIN_SIZE_Y) lprc->top = lprc->bottom - MIN_SIZE_Y ;
   break ;

   case WM_COMMAND:
      retcd = TRUE ;
      switch( LOWORD( wparam ) )
      {
      case DLG_B_INFO:
         MessageBox(window_handle,
                    "   Find set of resistors/capacitors\r"
                    "Version 0.93, " __DATE__ ", " __TIME__ "\r"
                    "                dk8pp@ish.de\r",
                    "Program Info", MB_OK) ;
      break ;

      case RAD_COMPR:
         CheckRadioButton(window_handle,RAD_COMPR,RAD_COMPRN,RAD_COMPR) ;
         icons   = LoadImage(MyInstance, "IC_RS",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconss  = LoadImage(MyInstance, "IC_RSS",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconsss = LoadImage(MyInstance, "IC_RSSS", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconpp  = LoadImage(MyInstance, "IC_RPP",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconspp = LoadImage(MyInstance, "IC_RSPP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconppp = LoadImage(MyInstance, "IC_RPPP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconssp = LoadImage(MyInstance, "IC_RSSP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         SendMessage(GetDlgItem(MainWindow, ICON_S),   STM_SETIMAGE, IMAGE_ICON, (LPARAM) icons  ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SS),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconss ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSS), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconsss); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PP),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconpp ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconspp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconppp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconssp); /* set icon */
         component = R ;
         EnableControls(TRUE) ; /* enable controls */
         EnableWindow(GetDlgItem(MainWindow, DLG_TARGET2), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMIN), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_ILOAD), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMAX), FALSE) ;
         SetDlgItemText(MainWindow, DLG_TARGETTEXT, "Target") ;
      break ;

      case RAD_COMPC:
         CheckRadioButton(window_handle,RAD_COMPR,RAD_COMPRN,RAD_COMPC) ;
         icons   = LoadImage(MyInstance, "IC_CS",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconss  = LoadImage(MyInstance, "IC_CSS",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconsss = LoadImage(MyInstance, "IC_CSSS", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconpp  = LoadImage(MyInstance, "IC_CPP",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconspp = LoadImage(MyInstance, "IC_CSPP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconppp = LoadImage(MyInstance, "IC_CPPP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconssp = LoadImage(MyInstance, "IC_CSSP", IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         SendMessage(GetDlgItem(MainWindow, ICON_S),   STM_SETIMAGE, IMAGE_ICON, (LPARAM) icons  ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SS),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconss ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSS), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconsss); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PP),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconpp ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconspp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconppp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconssp); /* set icon */
         component = C ;
         EnableControls(TRUE) ; /* enable controls */
         EnableWindow(GetDlgItem(MainWindow, DLG_TARGET2), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMIN), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_ILOAD), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMAX), FALSE) ;
         SetDlgItemText(MainWindow, DLG_TARGETTEXT, "Target") ;
      break ;

      case RAD_COMPRN:
         CheckRadioButton(window_handle,RAD_COMPR,RAD_COMPRN,RAD_COMPRN) ;
         icons   = LoadImage(MyInstance, "IC_RNSS",    IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconss  = LoadImage(MyInstance, "IC_RNSPP",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconsss = LoadImage(MyInstance, "IC_RNPPS",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconpp  = LoadImage(MyInstance, "IC_RNSSS1",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconspp = LoadImage(MyInstance, "IC_EMPTY",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconppp = LoadImage(MyInstance, "IC_RNSSS2",  IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         iconssp = LoadImage(MyInstance, "IC_EMPTY",   IMAGE_ICON, 0, 0, LR_DEFAULTSIZE) ;
         SendMessage(GetDlgItem(MainWindow, ICON_S),   STM_SETIMAGE, IMAGE_ICON, (LPARAM) icons  ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SS),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconss ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSS), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconsss); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PP),  STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconpp ); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconspp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_PPP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconppp); /* set icon */
         SendMessage(GetDlgItem(MainWindow, ICON_SSP), STM_SETIMAGE, IMAGE_ICON, (LPARAM) iconssp); /* set icon */
         CheckDlgButton(window_handle, DLG_TOPOS, BST_CHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOSS, BST_UNCHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOSSS, BST_UNCHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOPP, BST_UNCHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOSPP, BST_UNCHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOPPP, BST_UNCHECKED) ;
         CheckDlgButton(window_handle, DLG_TOPOSSP, BST_UNCHECKED) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_MAXDEVIATION), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOS), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPP), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSS), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPPP), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSS), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSPP), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSP), FALSE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_TARGET2), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMIN), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_ILOAD), TRUE) ;
         EnableWindow(GetDlgItem(MainWindow, DLG_IMAX), TRUE) ;
         SetDlgItemText(MainWindow, DLG_TARGETTEXT, "U1") ;
         component = RN ;
      break ;

      case DLG_MAXERROR:
         GetWindowText(GetDlgItem(window_handle, DLG_MAXERROR), buffer, sizeof(buffer)) ;
         MAXERR_error = TRUE ;
         if (get_double_value(buffer, &input_value) && (input_value > 0))
         {
            maxerror = input_value ;
            MAXERR_error = FALSE ;
         }
      break ;

      case DLG_MAXDEVIATION:
         GetWindowText(GetDlgItem(window_handle, DLG_MAXDEVIATION), buffer, sizeof(buffer)) ;
         MAXDEVIATION_error = TRUE ;
        if (get_double_value(buffer, &input_value) && (input_value > 0))
         {
            maxdeviation = input_value ;
            MAXDEVIATION_error = FALSE ;
         }
      break ;

      case DLG_SUSPEND:
         if (hThread != NULL)
         {
            if (!suspended)
            {
               SuspendThread(hThread) ;
               suspended = TRUE ;
               SetDlgItemText(MainWindow, DLG_SUSPEND, "Resume") ;
               LoadString(MyInstance, STR_HALTED, strng, MAX_LOADSTRING);
               SetDlgItemText(MainWindow, DLG_STATUS, strng) ;
            }
            else
            {
               ResumeThread(hThread) ;
               suspended = FALSE ;
               SetDlgItemText(MainWindow, DLG_SUSPEND, "Suspend") ;
               LoadString(MyInstance, STR_WORKING, strng, MAX_LOADSTRING);
               SetDlgItemText(MainWindow, DLG_STATUS, strng) ;
            }
         }
      break ;

      case DLG_START:
/* Start */
         fill_table() ; /* generate table with all component values */
         if (hThread != NULL)
         { /* Thread already started. Abort */
            abort_thread = TRUE ;
            LoadString(MyInstance, STR_ABORTING, strng, MAX_LOADSTRING);
            SetDlgItemText(MainWindow, DLG_STATUS, strng) ;
         }
         else
         {
/* invalidate solutions */
            for (i=0; i<NOS; i++) sol[i].valid = FALSE ;
            counter = 0 ;
/* clear output window */
            msg_hwnd = GetDlgItem(MainWindow, DLG_RESULT) ;
            SendMessage(msg_hwnd, EM_SETSEL, 0, -1); /* clear window */
            SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) FALSE, (WPARAM) "") ;
            abort_thread = FALSE ;

/* get target value */
            GetWindowText(GetDlgItem(window_handle, DLG_TARGET), buffer, sizeof(buffer)) ;
            TARGET_error = TRUE ;
            if (get_double_value(buffer, &input_value)
               && (input_value > 0)
               && (input_value < MAX_TARGET))
            {
               target = input_value ;
               TARGET_error = FALSE ;
            }

            if (component == RN)
            {
/* get target2 value */
               GetWindowText(GetDlgItem(window_handle, DLG_TARGET2), buffer, sizeof(buffer)) ;
               if (get_double_value(buffer, &input_value)
                  && (input_value > 0))
               {
                  target2 = input_value ;
               }
               else TARGET_error = TRUE ;

/* get Imin value */
               GetWindowText(GetDlgItem(window_handle, DLG_IMIN), buffer, sizeof(buffer)) ;
               if (strlen(buffer) > 0)
               {
                  if (get_double_value(buffer, &input_value)
                     && (input_value >= 0))
                  {
                     imin = input_value ;
                     valid_imin = TRUE ;
                  }
                  else TARGET_error = TRUE ;
               }
               else valid_imin = FALSE ;

/* get Imax value */
               GetWindowText(GetDlgItem(window_handle, DLG_IMAX), buffer, sizeof(buffer)) ;
               if (strlen(buffer) > 0)
               {
                  if (get_double_value(buffer, &input_value)
                     && (input_value >= 0))
                  {
                     imax = input_value ;
                     valid_imax = TRUE ;
                  }
                  else TARGET_error = TRUE ;
               }
               else valid_imax = FALSE ;

/* get Iload value */
               GetWindowText(GetDlgItem(window_handle, DLG_ILOAD), buffer, sizeof(buffer)) ;
               if (strlen(buffer) > 0)
               {
                  if (get_double_value(buffer, &input_value))
                     iload = input_value ;
                  else TARGET_error = TRUE ;
               }
               else iload = 0 ;
            }

            LoadString(MyInstance, STR_EMPTY, strng, MAX_LOADSTRING);
            if (TARGET_error)
            {
               LoadString(MyInstance, STR_INVALIDTARGET, strng, MAX_LOADSTRING);
            }

            if ((!TARGET_error) && (target2 > target) && (component == RN))
            {
               LoadString(MyInstance, STR_TARGETVAL, strng, MAX_LOADSTRING);
               TARGET_error = TRUE ;
            }
            SetDlgItemText(MainWindow, DLG_STATUS, strng) ;

            if (!TARGET_error)
            {
               topos   = IsDlgButtonChecked(window_handle, DLG_TOPOS) ;
               topopp  = IsDlgButtonChecked(window_handle, DLG_TOPOPP) ;
               toposs  = IsDlgButtonChecked(window_handle, DLG_TOPOSS) ;
               topoppp = IsDlgButtonChecked(window_handle, DLG_TOPOPPP) ;
               toposss = IsDlgButtonChecked(window_handle, DLG_TOPOSSS) ;
               topospp = IsDlgButtonChecked(window_handle, DLG_TOPOSPP) ;
               topossp = IsDlgButtonChecked(window_handle, DLG_TOPOSSP) ;

               SaveControls(status) ;  /* save status of controls */
               EnableControls(FALSE) ; /* disable some critical controls during execution */

               hThread = CreateThread(
                   NULL,                        // no security attributes
                   0,                           // use default stack size
                   FindResults,                 // thread function
                   &dwThrdParam,                // argument to thread function
                   0,                           // use default creation flags
                   &dwThreadId);                // returns the thread identifier

              // Check the return value for success.
               if (hThread == NULL)
                  LoadString(MyInstance, STR_THREADSTARTERROR, strng, MAX_LOADSTRING);
               else
               {
                  LoadString(MyInstance, STR_WORKING, strng, MAX_LOADSTRING);
                  suspended = FALSE ;
                  start_time = clock() ;
               }
               SetDlgItemText(MainWindow, DLG_STATUS, strng) ;
               SetDlgItemText(MainWindow, DLG_START, "Abort") ;
            }
         }
      break ;

      case IDABORT:
      case IDCANCEL:
         EndDialog(window_handle, 55);         /* Exits the dialog box */
      break ;

      default:
         retcd = FALSE ;
      break ;
      }
   }
   return(retcd);
} /* ParSerProc */


/*
+-----------------------------------------------------------------------+
|       get_coords                                                      |
+-----------------------------------------------------------------------+
|                                                                       |
| Get the coordinates of the controls, which are defined in the         |
| coordinates structure.                                                |
|                                                                       |
+-----------------------------------------------------------------------+
*/
void get_coords(COORDS *coordinates, _WORD size)
{
RECT rect ;
POINT pntlo ;
POINT pntru ;
_WORD i ;
   for (i=1; i<size; i++) /* skip first entry; reserved for main dialog */
   {
      GetWindowRect(GetDlgItem(MainWindow, coordinates[i].control), &rect) ;
      pntlo.x = rect.left ;
      pntlo.y = rect.top ;
      pntru.x = rect.right ;
      pntru.y = rect.bottom ;
      ScreenToClient(MainWindow, &pntlo) ;
      ScreenToClient(MainWindow, &pntru) ;
      coordinates[i].left = pntlo.x ;
      coordinates[i].top = pntlo.y ;
      coordinates[i].right = pntru.x ;
      coordinates[i].bottom = pntru.y ;
   }
}


/*
+-----------------------------------------------------------------------+
|       resize_controls                                                 |
+-----------------------------------------------------------------------+
|                                                                       |
| This function is called as a response to the WM_SIZE message.         |
| It resizes the controls of a dialog.                                  |
|                                                                       |
+-----------------------------------------------------------------------+
*/
void resize_controls(COORDS *coordinates, _WORD size)
{
RECT rect ;
_WORD i, hsize, vsize, hsize_org, vsize_org, addx, addy, movx, movy ;
// get current size of dialog
   GetWindowRect(MainWindow, &rect) ;
   hsize = rect.right - rect.left ;
   vsize = rect.bottom - rect.top ;
   hsize_org = coordinates[0].right-coordinates[0].left ;
   vsize_org = coordinates[0].bottom-coordinates[0].top ;

   for (i=1; i<size; i++)
   {
      if (((coordinates[i].flags)&RESIZE_X)!=0) addx = hsize-hsize_org ; else addx = 0 ;
      if (((coordinates[i].flags)&RESIZE_Y)!=0) addy = vsize-vsize_org ; else addy = 0 ;
      if (((coordinates[i].flags)&MOVE_X)!=0) movx = hsize-hsize_org ; else movx = 0 ;
      if (((coordinates[i].flags)&MOVE_Y)!=0) movy = vsize-vsize_org ; else movy = 0 ;
      MoveWindow(
      GetDlgItem(MainWindow, coordinates[i].control),   // handle of window
      coordinates[i].left+movx,                         // horizontal position
      coordinates[i].top+movy,                          // vertical position
      coordinates[i].right-coordinates[i].left+addx,    // width
      coordinates[i].bottom-coordinates[i].top+addy,    // height
      TRUE);                                            // repaint flag
   }
}


/*
*************************************************************************
*       void EnableControls(_BOOL enable)                               *
*************************************************************************
*                                                                       *
* Enable or disable some critical controls during execution of thread.  *
* Control gets enabled if enable is TRUE, otherwise it gets disabled.   *
*                                                                       *
*************************************************************************
*/
void EnableControls(_BOOL enable)
{
   EnableWindow(GetDlgItem(MainWindow, DLG_E6), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E12), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E24), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E48), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E96), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E192), enable) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPR), enable) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPC), enable) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPRN), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TARGET), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TARGET2), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_IMIN), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_IMAX), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_ILOAD), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_MAXERROR), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_MAXDEVIATION), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOS), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPP), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSS), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPPP), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSS), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSPP), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSP), enable) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_SUSPEND), !enable) ;
}

/*
*************************************************************************
*       void SaveControls(_BOOL *status)                                *
*************************************************************************
*                                                                       *
* Save enabled/disabled status of some controls.                        *
*                                                                       *
*************************************************************************
*/
void SaveControls(_BOOL *status)
{
   status[ 0] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E6)) ;
   status[ 1] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E12)) ;
   status[ 2] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E24)) ;
   status[ 3] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E48)) ;
   status[ 4] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E96)) ;
   status[ 5] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_E192)) ;
   status[ 6] = IsWindowEnabled(GetDlgItem(MainWindow, RAD_COMPR)) ;
   status[ 7] = IsWindowEnabled(GetDlgItem(MainWindow, RAD_COMPC)) ;
   status[ 8] = IsWindowEnabled(GetDlgItem(MainWindow, RAD_COMPRN)) ;
   status[ 9] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TARGET)) ;
   status[10] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TARGET2)) ;
   status[11] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_IMIN)) ;
   status[12] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_IMAX)) ;
   status[13] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_ILOAD)) ;
   status[14] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_MAXERROR)) ;
   status[15] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_MAXDEVIATION)) ;
   status[16] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOS)) ;
   status[17] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOPP)) ;
   status[18] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOSS)) ;
   status[19] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOPPP)) ;
   status[20] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOSSS)) ;
   status[21] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOSPP)) ;
   status[22] = IsWindowEnabled(GetDlgItem(MainWindow, DLG_TOPOSSP)) ;
}

/*
*************************************************************************
*       void RestoreControls(_BOOL *status)                             *
*************************************************************************
*                                                                       *
* Restore enabled/disabled status of some controls.                     *
*                                                                       *
*************************************************************************
*/
void RestoreControls(_BOOL *status)
{
   EnableWindow(GetDlgItem(MainWindow, DLG_E6),           status[ 0]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E12),          status[ 1]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E24),          status[ 2]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E48),          status[ 3]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E96),          status[ 4]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_E192),         status[ 5]) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPR),        status[ 6]) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPC),        status[ 7]) ;
   EnableWindow(GetDlgItem(MainWindow, RAD_COMPRN),       status[ 8]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TARGET),       status[ 9]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TARGET2),      status[10]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_IMIN),         status[11]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_IMAX),         status[12]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_ILOAD),        status[13]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_MAXERROR),     status[14]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_MAXDEVIATION), status[15]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOS),        status[16]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPP),       status[17]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSS),       status[18]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOPPP),      status[19]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSS),      status[20]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSPP),      status[21]) ;
   EnableWindow(GetDlgItem(MainWindow, DLG_TOPOSSP),      status[22]) ;
}

/*
*************************************************************************
*       void print_results(void)                                        *
*************************************************************************
*                                                                       *
* Display the results.                                                  *
*                                                                       *
*************************************************************************
*/
void print_results(void)
{
HWND msg_hwnd ;
_WORD iLen, i, j ;
BOOL found = FALSE ;
_BYTE str1[10], str2[10], str3[10], res[10] ;
   msg_hwnd = GetDlgItem(MainWindow, DLG_RESULT) ;
   iLen = GetWindowTextLength(msg_hwnd) ;
   SendMessage(msg_hwnd, EM_SETSEL, 0, -1); /* clear window */
   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) FALSE, (WPARAM) "") ;
   do
   {
      j = 0 ;
      found = FALSE ;
      for (i=0; i<NOS; i++)
      {
         if ((sol[i].valid) && ((!sol[j].valid) || ((sol[j].err) > sol[i].err)))
         {
            j = i ;
            found = TRUE ;
         }
      }
      found = found || (sol[j].valid) ;
      if (found)
      {
         sol[j].valid = FALSE ;
         pval(str1, sol[j].val1) ;
         pval(str2, sol[j].val2) ;
         pval(str3, sol[j].val3) ;
         pval(res, sol[j].res) ;
         switch (sol[j].conf) {
         case S:
            sprintf(strng, "%s (=%1.2lf%% error)\r\r\n",
                           str1, 100*sol[j].err/sol[j].target) ;
            break ;

         case SS:
            sprintf(strng, "%s - %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case PP:
            sprintf(strng, "%s || %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case PPP:
            sprintf(strng, "%s || %s || %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case SPP:
            sprintf(strng, "%s - %s || %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case SSS:
            sprintf(strng, "%s - %s - %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case SSP:
            sprintf(strng, "(%s - %s) || %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case SSS1:
            sprintf(strng, "%s - (%s - %s) = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case SSS2:
            sprintf(strng, "(%s - %s) - %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         case PPS:
            sprintf(strng, "%s || %s - %s = %s (=%1.2lf%% error)\r\r\n",
                           str1, str2, str3, res,
                           100*sol[j].err/sol[j].target) ;
            break ;

         }
         SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) FALSE, (WPARAM) strng) ;
      }
   } while (found) ;
   SendMessage(msg_hwnd, EM_SETSEL, 0, 0); /* position to top left */
   SendMessage(msg_hwnd, EM_SCROLLCARET, 0, 0); /* scroll caret into view */
}


/*
*************************************************************************
*       _BYTE *pval(_BYTE *strng, _DOUBLE val)                          *
*************************************************************************
*                                                                       *
* Format an output value. Use suffixes 'k' or 'M' where appropriate.    *
*                                                                       *
*************************************************************************
*/
_BYTE *pval(_BYTE *strng, _DOUBLE val)
{
_BYTE *suffix ;
_BYTE *form ;
   suffix = "" ;

   if (val < 1)
   {
      val *= 1000 ;
      suffix = "m" ;
   }

   if (val < 1)
   {
      val *= 1000 ;
      suffix = "" ;
   }

   if (val < 1)
   {
      val *= 1000 ;
      suffix = "n" ;
   }

   if (val < 1)
   {
      val *= 1000 ;
      suffix = "p" ;
   }

   if (val >= 1000)
   {
      val /= 1000 ;
      suffix = "k" ;
   }

   if (val >= 1000)
   {
      val /= 1000 ;
      suffix = "M" ;
   }

   if (val - (_DOUBLE)((_WORD) val) > 0.01)
      form = "%1.2lf%s" ;
   else
      form = "%1.0lf%s" ;

   sprintf(strng, form, val, suffix) ;

   return(strng) ;
}

/*
+-----------------------------------------------------------------------+
|       BOOL get_double_value(_BYTE *buffer, _DOUBLE *value)            |
+-----------------------------------------------------------------------+
|                                                                       |
| Decode an integer value with optional suffix (k,M,G...)               |
| Decimal points are supported, but result is always truncated to       |
| signed long long.                                                     |
|                                                                       |
+-----------------------------------------------------------------------+
*/
BOOL get_double_value(_BYTE *buffer, _DOUBLE *value)
{
_DOUBLE mydouble ;
_DOUBLE mult = 1.0 ;
_BYTE mystrng[16] ;
_LONG size ;
    size = sscanf(buffer, "%lf%s", &mydouble, &mystrng) ;
    if (size == 2)
    {
        switch (mystrng[0]) {
        case 'T': mult = 1e12 ; break ;
        case 'G': mult = 1e9 ; break ;
        case 'M': mult = 1e6 ; break ;
        case 'k': mult = 1e3 ; break ;
        case 'm': mult = 1e-3 ; break ;
        case 'u': mult = 1e-6 ; break ;
        case '': mult = 1e-6 ; break ;
        case 'n': mult = 1e-9 ; break ;
        case 'p': mult = 1e-12 ; break ;
        case '%': mult = 1e-2 ; break ;
        default : mult = 1.0 ; }
    }
    *value = mydouble * mult ;
    return((size == 1) || (size == 2)) ;
}


/*
+-----------------------------------------------------------------------+
|       _WORD mix_table(_DOUBLE *target, _DOUBLE *source1, _WORD size1, |
|                       _DOUBLE *source2, _WORD size2)                  |
+-----------------------------------------------------------------------+
|                                                                       |
| Mix two tables and return the size of the new table.                  |
|                                                                       |
+-----------------------------------------------------------------------+
*/
_WORD mix_table(_DOUBLE *target, _DOUBLE *source1, _WORD size1,
                                 _DOUBLE *source2, _WORD size2)
{
_WORD size, i, j, k ;
_DOUBLE v1, last ;
   size = 0 ;
   i = 0 ;
   j = 0 ;
   k = 0 ;
   last = 0 ;
   while ((i<size1) || (j<size2))
   {
      if (i<size1)
      {
         v1 = source1[i] ;
         i++ ;
      }
      else
         v1 = -1 ;
      while ((j<size2) && (((v1 >=0) && (source2[j] <= v1)) || (v1 < 0)))
      {
         target[size] = source2[j] ;
         last = source2[j] ;
         size++ ;
         j++ ;
      }
      if ((v1 > 0) && (v1 != last))
      {
         target[size] = v1 ;
         size++ ;
      }
   }
   return(size) ;
}

/*
+-----------------------------------------------------------------------+
|       void fill_table(void)                                           |
+-----------------------------------------------------------------------+
|                                                                       |
| Fill the table with all decades of currently selected tables.         |
|                                                                       |
+-----------------------------------------------------------------------+
*/
void fill_table(void)
{
_WORD i, j, dec ;
_DOUBLE mul ;
_WORD size ;
_DOUBLE table1[(sizeof(e192)+sizeof(e24))/sizeof(e192[0])] ; /* e24 contains some unique values */
_DOUBLE table2[(sizeof(e192)+sizeof(e24))/sizeof(e192[0])] ; /* e24 contains some unique values */
_DOUBLE *source, *target, *temp ;
   size = 0 ;
   target = table1 ;
   source = table2 ;
   if (IsDlgButtonChecked(MainWindow, DLG_E6))
   {
      size = mix_table(target, e6, sizeof(e6)/sizeof(e6[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   if (IsDlgButtonChecked(MainWindow, DLG_E12))
   {
      size = mix_table(target, e12, sizeof(e12)/sizeof(e12[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   if (IsDlgButtonChecked(MainWindow, DLG_E24))
   {
      size = mix_table(target, e24, sizeof(e24)/sizeof(e24[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   if (IsDlgButtonChecked(MainWindow, DLG_E48))
   {
      size = mix_table(target, e48, sizeof(e48)/sizeof(e48[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   if (IsDlgButtonChecked(MainWindow, DLG_E96))
   {
      size = mix_table(target, e96, sizeof(e96)/sizeof(e96[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   if (IsDlgButtonChecked(MainWindow, DLG_E192))
   {
      size = mix_table(target, e192, sizeof(e192)/sizeof(e192[0]), source, size) ;
      temp = target ;
      target = source ;
      source = temp ;
   }
   mul = ((component == R) || (component == RN)) ? 1e-3 : 1e-14 ;
   dec = ((component == R) || (component == RN)) ? RDECADES : CDECADES ;
   tablesize = dec * size ;
   for (j=0; j<dec; j++)
   {
       for (i=0; i<size; i++)
       {
          table[i+size*j] = source[i] * mul ;
       }
       mul *= 10.0 ;
  }
}


/*
+-----------------------------------------------------------------------+
|       void enqueue(_DOUBLE target, _DOUBLE res, CONF conf,            |
|                    _DOUBLE val1, _DOUBLE val2, _DOUBLE val3)          |
+-----------------------------------------------------------------------+
|                                                                       |
| Enqueue values if error is below maximum error of current entries     |
| and lower than the maximum permitted error.                           |
|                                                                       |
+-----------------------------------------------------------------------+
*/
void enqueue(_DOUBLE target, _DOUBLE res, CONF conf,
             _DOUBLE val1, _DOUBLE val2, _DOUBLE val3)
{
_WORD i, j ;
BOOL found ;
_DOUBLE error ; /* absolute error */
_DOUBLE deviation ; /* deviation between the individual components */
_DOUBLE maxf, minf ;
   found = FALSE ;
   error = fabs(res-target) ;
   if (conf == S) deviation = 0 ;
   if ((conf == SS) || (conf == PP))
      deviation = (val1>val2) ? val1/val2 : val2/val1 ;
   if ((conf == SSS) || (conf == SPP) || (conf == PPP) || (conf == SSP) || (conf == PPS) || (conf == SSS1) || (conf == SSS2))
   {
      maxf = (val1 > val2) ? val1 : val2 ;
      maxf = (maxf > val3) ? maxf : val3 ;
      minf = (val1 < val2) ? val1 : val2 ;
      minf = (minf < val3) ? minf : val3 ;
      deviation = maxf / minf ;
   }
   if (error/target <= maxerror)
   {
      if ((MAXDEVIATION_error) || (deviation <= maxdeviation))
      {
         j = 0 ;
         for (i=0; (i<NOS) && (!found); i++) found = (!sol[i].valid) ;
         j = i - 1 ;
         if (!found)
         /* table is full. Seek entry with max error */
            for (i=0; i<NOS; i++) if (sol[i].err > sol[j].err) j = i ;
         if ((sol[j].err > error) || found)
         { /* enque values */
            sol[j].target = target ;
            sol[j].val1 = val1 ;
            sol[j].val2 = val2 ;
            sol[j].val3 = val3 ;
            sol[j].conf = conf ;
            sol[j].res = res ;
            sol[j].err  = error ;
            sol[j].valid = TRUE ;
         }
         counter++ ;
         SendMessage(MainWindow, WM_USER, USER_PRINTCOUNT, 0) ;
      }
   }
}

/*
+-----------------------------------------------------------------------+
|       FindResults(LPVOID lpParam)                                     |
+-----------------------------------------------------------------------+
|                                                                       |
| This thread searches for the possible results. The parameter is       |
| unused; we use only global variables.                                 |
|                                                                       |
| The user can abort this thread by setting the global variable         |
| abort_thread to TRUE.                                                 |
|                                                                       |
+-----------------------------------------------------------------------+
*/
DWORD WINAPI FindResults(LPVOID lpParam)
{
_DOUBLE limit1, limit2, llimit, ulimit, llimitres, ulimitres, itotal, rges ;
_WORD i, j, k ;
CONF cfg ;
   ulimit = target * (1 + maxerror) ; /* upper limit */
   llimit = target / (1 + maxerror) ; /* lower limit */
// check S configuration
   if ((topos) && ((component == R) || (component == C)))
   {
      limit1 = target * (1 + maxerror) ;
      for (i = 0; (i < tablesize) && (table[i] < llimit); i++) ;
      if (i > 0) i-- ;
      for ( ; (i < tablesize) && (table[i] <= limit1); i++)
         enqueue(target, table[i], S, table[i], 0, 0) ;
   }

// check PP configuration
   if ( (((component == R) && (topopp))
      || ((component == C) && (toposs))) && (!abort_thread))
   {
      if (component == R) cfg = PP ;
      if (component == C) cfg = SS ;
      for (i = 0; (i < tablesize) && (table[i] < llimit); i++) ;
      if (i > 0) i-- ;
      for ( ; (i < tablesize) && (!abort_thread); i++)
      {
         limit1 = table[i] * maxdeviation ;
         for (j = i; (j < tablesize) && (table[j] <= limit1); j++)
         {
            res = 1/((1/table[i])+(1/table[j])) ;
            enqueue(target, res, cfg, table[i], table[j], 0) ;
         }
      }
   }

// check SS configuration
   if ( (((component == R) && (toposs))
      || ((component == C) && (topopp))) && (!abort_thread))
   {
      if (component == R) cfg = SS ;
      if (component == C) cfg = PP ;
      for (i = 0; (i < tablesize) && (table[i] <= ulimit) && (!abort_thread); i++)
      {
         limit1 = min(ulimit, table[i] * maxdeviation) ;
         for (j = i; (j < tablesize) && (table[j] <= limit1); j++)
         {
            res = table[i] + table[j] ;
            enqueue(target, res, cfg, table[i], table[j], 0) ;
         }
      }
   }

// check PPP configuration
   if ( (((component == R) && (topoppp))
      || ((component == C) && (toposss))) && (!abort_thread))
   {
      if (component == R) cfg = PPP ;
      if (component == C) cfg = SSS ;
      for (i = 0; (i < tablesize) && (table[i] < llimit); i++) ;
      if (i > 0) i-- ;
      for ( ; (i < tablesize) && (!abort_thread); i++)
      {
         limit1 = table[i] * maxdeviation ;
         for (j = i; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(table[i],table[j]) * maxdeviation ;
            for (k = j; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = 1/((1/table[i])+(1/table[j])+(1/table[k])) ;
               enqueue(target, res, cfg, table[i], table[j], table[k]) ;
            }
         }
      }
   }

// check SPP configuration for resistors
   if ((component == R) && (topospp) && (!abort_thread))
   {
      cfg = SPP ;
      for (i = 0; (i < tablesize) && (table[i] <= ulimit) && (!abort_thread); i++)
      {
         limit1 = table[i] * maxdeviation ;
         for (j = 0; (j < tablesize) && (table[j] < (llimit-table[i])); j++) ;
         if (j > 0) j-- ;
         for ( ; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(table[i],table[j]) * maxdeviation ;
            for (k = j ; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = table[i] + 1/((1/table[j])+(1/table[k])) ;
               enqueue(target, res, cfg, table[i], table[j], table[k]) ;
            }
         }
      }
   }

// check SPP configuration for capacitors
   if ((component == C) && (topospp) && (!abort_thread))
   {
      cfg = SPP ;
      for (i = 0; (i < tablesize) && (table[i] < llimit); i++) ;
      if (i > 0) i-- ;
      for ( ; (i < tablesize) && (!abort_thread); i++)
      {
         limit1 = table[i] * maxdeviation ;
         for (j = 0; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(table[i],table[j]) * maxdeviation ;
            for (k = j; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = 1/(1/table[i] + 1/(table[j]+table[k])) ;
               enqueue(target, res, cfg, table[i], table[j], table[k]) ;
            }
         }
      }
   }

// check SSP configuration for resistors
   if ((component == R) && (topossp) && (!abort_thread))
   {
      cfg = SSP ;
      for (i = 0; (i < tablesize) && (table[i] < llimit); i++) ;
      if (i > 0) i-- ;
      for ( ; (i < tablesize) && (!abort_thread); i++)
      {
         ulimitres = 1/((1/ulimit) - (1/table[i])) ;
         llimitres = 1/((1/llimit) - (1/table[i])) ;
         limit1 = min(llimitres, table[i] * maxdeviation) ;
         for (j = 0; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(llimitres, min(table[i],table[j]) * maxdeviation) ;
            for (k = j; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = 1/((1/table[i])+1/(table[j] + table[k])) ;
               enqueue(target, res, cfg, table[j], table[k], table[i]) ;
            }
         }
      }
   }

// check SSP configuration for capacitors
   if ((component == C) && (topossp) && (!abort_thread))
   {
      cfg = SSP ;
      for (i = 0; (i < tablesize) && (table[i] < ulimit) && (!abort_thread); i++)
      {
         limit1 = table[i] * maxdeviation ;
         for (j = 0; (j < tablesize) && (table[j] < (llimit-table[i])); j++) ;
         if (j > 0) j-- ;
         for ( ; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(table[i],table[j]) * maxdeviation ;
            for (k = j; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = table[i] + 1/(1/table[j] + 1/table[k]) ;
               enqueue(target, res, cfg, table[j], table[k], table[i]) ;
            }
         }
      }
   }

// check SSS configuration
   if ( (((component == R) && (toposss))
      || ((component == C) && (topoppp))) && (!abort_thread))
   {
      if (component == R) cfg = SSS ;
      if (component == C) cfg = PPP ;
      for (i = 0; (i < tablesize) && (table[i] <= ulimit) && (!abort_thread); i++)
      {
         limit1 = min(ulimit, table[i] * maxdeviation) ;
         for (j = i; (j < tablesize) && (table[j] <= limit1); j++)
         {
            limit2 = min(ulimit, min(table[i],table[j]) * maxdeviation) ;
            for (k = j; (k < tablesize) && (table[k] <= limit2); k++)
            {
               res = table[i]+table[j]+table[k] ;
               enqueue(target, res, cfg, table[i], table[j], table[k]) ;
            }
         }
     }
   }

// check RN configuration
   if (component == RN)
   {
      if (topos)
      {
         for (i = 0; (i < tablesize) && (!abort_thread); i++)
         {
            for (j = 0; j < tablesize; j++)
            {
               itotal = target / (table[i] + table[j]) ;
               if ((!valid_imin || (itotal >= imin)) && (!valid_imax || (itotal <= imax)))
               {
                  res = (target - iload * table[i]) * table[j] / (table[i] + table[j]) ;
                  enqueue(target2, res, SS, table[i], table[j], 0) ;
               }
            }
         }
      }


      if (toposs)
      {
         for (i = 0; (i < tablesize) && (!abort_thread); i++)
         {
            for (j = 0; j < tablesize; j++)
            {
               for (k = j; k < tablesize; k++)
               {
                  rges = (1 / ((1 / table[j]) + (1 / table[k]))) ; /* R2 || R3 */
                  itotal = target / (table[i] + rges) ;
                  if ((!valid_imin || (itotal >= imin)) && (!valid_imax || (itotal <= imax)))
                  {
                     res = (target - iload * table[i]) * rges / (table[i] + rges) ;
                     enqueue(target2, res, SPP, table[i], table[j], table[k]) ;
                  }
               }
            }
         }
      }

      if (topopp)
      {
         for (i = 0; (i < tablesize) && (!abort_thread); i++)
         {
            for (j = 0; j < tablesize; j++)
            {
               for (k = j; k < tablesize; k++)
               {
                  rges = table[j] + table[k] ; /* R2 - R3 */
                  itotal = target / (table[i] + rges) ;
                  if ((!valid_imin || (itotal >= imin)) && (!valid_imax || (itotal <= imax)))
                  {
                     res = (target - iload * table[i]) * rges / (table[i] + rges) ;
                     enqueue(target2, res, SSS1, table[i], table[j], table[k]) ;
                  }
               }
            }
         }
      }

      if (topoppp)
      {
         for (i = 0; (i < tablesize) && (!abort_thread); i++)
         {
            for (j = i; j < tablesize; j++)
            {
               for (k = 0; k < tablesize; k++)
               {
                  rges = table[i] + table[j] ; /* R1 - R2 */
                  itotal = target / (table[k] + rges) ;
                  if ((!valid_imin || (itotal >= imin)) && (!valid_imax || (itotal <= imax)))
                  {
                     res = (target - iload * rges) * table[k] / (table[k] + rges) ;
                     enqueue(target2, res, SSS2, table[i], table[j], table[k]) ;
                  }
               }
            }
         }
      }

      if (toposss)
      {
         for (i = 0; (i < tablesize) && (!abort_thread); i++)
         {
            for (j = i; j < tablesize; j++)
            {
               for (k = 0; k < tablesize; k++)
               {
                  rges = (1 / ((1 / table[i]) + (1 / table[j]))) ; /* R1 || R2 */
                  itotal = target / (table[k] + rges) ;
                  if ((!valid_imin || (itotal >= imin)) && (!valid_imax || (itotal <= imax)))
                  {
                     res = (target - iload * rges) * table[k] / (table[k] + rges) ;
                     enqueue(target2, res, PPS, table[i], table[j], table[k]) ;
                  }
               }
            }
         }
      }

   }

// Notify main window that the thread is finished
    SendMessage(MainWindow, WM_USER, USER_PRINTRESULTS, 0) ;

    return 0;
}


